"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.colorStatus = exports.errorTemplate = exports.Handler = void 0;
const pipeline_1 = require("@cardboardrobots/pipeline");
const console_style_1 = require("@cardboardrobots/console-style");
const context_1 = require("../context");
const directive_1 = require("../directive");
const Errors_1 = require("./Errors");
const LogLevel_1 = require("./LogLevel");
const DEFAULT_TEMPLATE = 'index';
const ERROR_TEMPLATE = 'error';
class Handler {
    constructor() {
        this.pipeline = new pipeline_1.Pipeline();
        this.errorPipeline = new pipeline_1.Pipeline();
        this.viewPipeline = new pipeline_1.Pipeline();
        this.logging = LogLevel_1.LogLevel.Errors;
        this.defaultTemplate = DEFAULT_TEMPLATE;
        this.callback = async (request, response) => {
            const context = new context_1.Context(request, response);
            try {
                const result = await this.pipeline.run(context, undefined);
                // If Headers have already sent, we cannot send.
                if (!context.response.headersSent && !context.response.writableEnded) {
                    this.send(context, result);
                }
            }
            catch (err) {
                const errorContext = context;
                const wrappedError = err instanceof Error ? err : new Error(err);
                errorContext.error = wrappedError;
                let errorStatus = 500;
                if (wrappedError instanceof Errors_1.SierraError) {
                    switch (wrappedError.message) {
                        case Errors_1.ErrorMessage.NotFound:
                            errorStatus = 404;
                            break;
                    }
                }
                if (this.errorPipeline.middlewares.length) {
                    try {
                        const result = await this.errorPipeline.run(errorContext, wrappedError);
                        switch (result.type) {
                            case pipeline_1.DirectiveType.Exit:
                            case pipeline_1.DirectiveType.End: {
                                const wrappedError = result.value instanceof Error
                                    ? result.value
                                    : new Error(result.value);
                                this.sendError(errorContext, directive_1.error(wrappedError, { status: errorStatus }));
                                break;
                            }
                            default:
                                this.send(errorContext, result);
                                break;
                        }
                    }
                    catch (err) {
                        const wrappedWrror = err instanceof Error ? err : Error(err);
                        errorContext.error = wrappedWrror;
                        this.sendError(errorContext, directive_1.error(wrappedWrror, { status: errorStatus }));
                    }
                }
                else {
                    this.sendError(errorContext, directive_1.error(wrappedError, { status: errorStatus }));
                }
            }
        };
    }
    sendJson(context, directive) {
        const { value, options } = directive;
        this.log(context, options.status);
        return this.write(context, JSON.stringify(value !== null && value !== void 0 ? value : null), options.status, Object.assign({ 'Content-Type': 'application/json' }, options.header));
    }
    sendRaw(context, directive) {
        var _a, _b, _c;
        const { value, options } = directive;
        this.log(context, options.status);
        if (typeof value === 'string') {
            return this.write(context, value, options.status, Object.assign({ 'Content-Type': (_a = options.contentType) !== null && _a !== void 0 ? _a : 'text/plain' }, options.header));
        }
        else if (Buffer.isBuffer(value)) {
            return this.write(context, value, options.status, Object.assign({ 'Content-Type': (_b = options.contentType) !== null && _b !== void 0 ? _b : 'octet-stream' }, options.header));
        }
        else {
            return this.write(context, JSON.stringify(value !== null && value !== void 0 ? value : null), options.status, Object.assign({ 'Content-Type': (_c = options.contentType) !== null && _c !== void 0 ? _c : 'application/json' }, options.header));
        }
    }
    sendText(context, directive) {
        const { value, options } = directive;
        this.log(context, options.status);
        let output;
        if (typeof value === 'object') {
            output = JSON.stringify(value !== null && value !== void 0 ? value : null);
        }
        else {
            output = `${value !== null && value !== void 0 ? value : ''}`;
        }
        return this.write(context, output, options.status, Object.assign({ 'Content-Type': 'text/plain' }, options.header));
    }
    async sendView(context, directive) {
        var _a, _b;
        const { options } = directive;
        try {
            context.template = (_b = (_a = options.template) !== null && _a !== void 0 ? _a : context.template) !== null && _b !== void 0 ? _b : this.defaultTemplate;
            const viewContext = context;
            viewContext.view = directive;
            const { value } = await this.viewPipeline.run(viewContext, directive.value);
            // Ensure output is a string
            if (typeof value !== 'string') {
                throw new Errors_1.NonStringViewError(value);
            }
            this.log(context, options.status);
            return this.write(context, value, options.status, Object.assign({ 'Content-Type': 'text/html' }, options.header));
        }
        catch (err) {
            this.log(context, 500);
            const result = await this.write(context, errorTemplate(err), 500, Object.assign({ 'Content-Type': 'text/html' }, options.header));
            if (this.logging >= LogLevel_1.LogLevel.Errors) {
                console.error(err);
            }
            return result;
        }
    }
    async sendAuto(context, directive) {
        const { accept } = context;
        if (this.viewPipeline.middlewares.length && accept && accept.indexOf('text/html') > -1) {
            const { value, options } = directive;
            this.sendView(context, directive_1.view(value, options));
        }
        else if (accept && accept.indexOf('application/json')) {
            this.sendJson(context, directive);
        }
        else {
            this.sendRaw(context, directive);
        }
    }
    async sendError(context, directive) {
        var _a;
        const { accept } = context.request.headers;
        const { value, options } = directive;
        const { status, header } = options;
        if (Math.floor(status / 100) === 5) {
            if (this.logging >= LogLevel_1.LogLevel.Errors) {
                console.error(value);
            }
        }
        try {
            if (this.viewPipeline.middlewares.length &&
                accept &&
                accept.indexOf('text/html') > -1) {
                // TODO: Fix sendView for Error
                await this.sendView(context, directive_1.view(value.message, { status, header, template: ERROR_TEMPLATE }));
            }
            else {
                this.sendJson(context, directive_1.json((_a = value.message) !== null && _a !== void 0 ? _a : value.name, { status, header }));
            }
        }
        catch (err) {
            if (this.logging >= LogLevel_1.LogLevel.Errors) {
                console.error(err);
            }
        }
    }
    send(context, directive) {
        if (directive instanceof directive_1.ResponseDirective) {
            if (directive instanceof directive_1.JsonDirective) {
                return this.sendJson(context, directive);
            }
            else if (directive instanceof directive_1.RawDirective) {
                return this.sendRaw(context, directive);
            }
            else if (directive instanceof directive_1.ViewDirective) {
                return this.sendView(context, directive);
            }
            else if (directive instanceof directive_1.AutoDirective) {
                return this.sendAuto(context, directive);
            }
            else if (directive instanceof directive_1.TextDirective) {
                return this.sendText(context, directive);
            }
            else if (directive instanceof directive_1.ErrorDirective) {
                return this.sendError(context, directive);
            }
            else {
                // TODO: Should this be Json or Auto?
                // This case should not happen
                return this.sendAuto(context, directive_1.auto(directive.value));
            }
        }
        else {
            // TODO: Should this be Json or Auto?
            return this.sendAuto(context, directive_1.auto(directive.value));
        }
    }
    log(context, status = 500) {
        if (this.logging >= LogLevel_1.LogLevel.Verbose) {
            console.log(context.request.method, context.request.url, colorStatus(status));
        }
    }
    use(middleware) {
        this.pipeline.use(middleware);
        return this;
    }
    useError(middlware) {
        return this.errorPipeline.use(middlware);
    }
    useView(middlware) {
        return this.viewPipeline.use(middlware);
    }
    async write(context, value, status, header) {
        const { response } = context;
        response.statusCode = status;
        Object.keys(header).forEach((headerName) => {
            const value = header[headerName];
            response.setHeader(headerName, value);
        });
        try {
            const result = await new Promise((resolve, reject) => {
                const result = response.write(value, (error) => {
                    if (error) {
                        reject(error);
                    }
                    else {
                        resolve(result);
                    }
                });
            });
            return result;
        }
        finally {
            await new Promise((resolve) => {
                response.end(() => resolve());
            });
        }
    }
}
exports.Handler = Handler;
function errorTemplate(error) {
    return `<!DOCTYPE html>
<html>
    <head>
        <title>Sierra Error</title>
    </head>
    <body>
        <h1>Sierra Error</h1>
        <pre><code>${error}</code></pre>
    </body>
</html>`;
}
exports.errorTemplate = errorTemplate;
function colorStatus(status) {
    switch (Math.floor(status / 100)) {
        case 1:
            return console_style_1.Color.white(status);
        case 2:
            return console_style_1.Color.green(status);
        case 3:
            return console_style_1.Color.blue(status);
        case 4:
            return console_style_1.Color.yellow(status);
        case 5:
            return console_style_1.Color.red(status);
        default:
            return console_style_1.Color.brightBlack(status);
    }
}
exports.colorStatus = colorStatus;
//# sourceMappingURL=Handler.js.map